"""Validation tests for :mod:`biomappings`."""

from __future__ import annotations

import getpass
import unittest
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from textwrap import dedent
from typing import ClassVar, TypeVar, cast

import bioregistry
from bioregistry import NormalizedNamableReference
from curies import NamableReference

from biomappings.resources import (
    CURATORS_PATH,
    SemanticMapping,
    _CuratedTuple,
    _PredictedTuple,
    load_mappings,
    load_predictions,
    mapping_sort_key,
)
from biomappings.resources.semapv import get_semapv_id_to_name
from biomappings.utils import (
    NEGATIVES_SSSOM_PATH,
    POSITIVES_SSSOM_PATH,
    UNSURE_SSSOM_PATH,
    get_canonical_tuple,
)

__all__ = [
    "IntegrityTestCase",
    "PathIntegrityTestCase",
]


SEMAPV_ID_TO_NAME = get_semapv_id_to_name()

TPL = TypeVar("TPL", _CuratedTuple, _PredictedTuple)
X = TypeVar("X")
Y = TypeVar("Y")


def _extract_redundant(counter: dict[X, list[Y]]) -> list[tuple[X, list[Y]]]:
    return [(key, values) for key, values in counter.items() if len(values) > 1]


def _locations_str(locations) -> str:
    return ", ".join(f"{label}:{line}" for label, line in locations)


class IntegrityTestCase(unittest.TestCase):
    """Data integrity tests."""

    mappings: list[SemanticMapping]
    predictions: list[SemanticMapping]
    incorrect: list[SemanticMapping]
    unsure: list[SemanticMapping]

    def _iter_groups(self) -> Iterable[tuple[str, int, SemanticMapping]]:
        for group, label in [
            (self.mappings, "positive"),
            (self.incorrect, "negative"),
            (self.predictions, "predictions"),
            (self.unsure, "unsure"),
        ]:
            for i, mapping in enumerate(group, start=2):
                yield label, i, mapping

    def test_mapping_justifications(self) -> None:
        """Test that the prediction type is pulled in properly."""
        for label, line, mapping in self._iter_groups():
            self.assertEqual(
                "semapv",
                mapping.mapping_justification.prefix,
                msg=f"[{label}] The 'mapping_justification' column should be annotated with semapv on line {line}",
            )
            self.assertIn(mapping.mapping_justification.identifier, SEMAPV_ID_TO_NAME)

    def test_prediction_not_manual(self) -> None:
        """Test that predicted mappings don't use manual mapping justification."""
        for _line, mapping in enumerate(self.predictions, start=2):
            self.assertNotEqual(
                "ManualMappingCuration",
                mapping.mapping_justification.identifier,
                msg="Prediction can not be annotated with manual curation",
            )

    def test_valid_curies(self) -> None:
        """Test that all mappings use canonical bioregistry prefixes."""
        for label, line, mapping in self._iter_groups():
            self.assert_valid(label, line, mapping.subject)
            self.assert_valid(label, line, mapping.predicate)
            self.assert_valid(label, line, mapping.object)
            self.assert_valid(label, line, mapping.mapping_justification)
            if mapping.author is not None:
                self.assert_valid(label, line, mapping.author)

    def assert_valid(self, label: str, line: int, reference: NamableReference) -> None:
        """Assert a reference is valid and normalized to the Bioregistry."""
        norm_prefix = bioregistry.normalize_prefix(reference.prefix)
        self.assertIsNotNone(
            norm_prefix, msg=f"Unknown prefix: {reference.prefix} on {label}:{line}"
        )
        self.assertEqual(
            norm_prefix,
            reference.prefix,
            msg=f"Non-normalized prefix: {reference.prefix} on {label}:{line}",
        )
        self.assertEqual(
            bioregistry.standardize_identifier(reference.prefix, reference.identifier),
            reference.identifier,
            msg=f"Invalid identifier: {reference.curie} on {label}:{line}",
        )

    def test_contributors(self) -> None:
        """Test all contributors have an entry in the curators.tsv file."""
        files = [
            (POSITIVES_SSSOM_PATH, self.mappings),
            (NEGATIVES_SSSOM_PATH, self.incorrect),
            (UNSURE_SSSOM_PATH, self.unsure),
        ]
        for path, mappings in files:
            for mapping in mappings:
                self.assertIsNotNone(mapping.author)
                author = cast(NormalizedNamableReference, mapping.author)
                self.assertEqual(
                    "orcid",
                    author.prefix,
                    msg=dedent(f"""
                    There are some curations that don't have the right metadata.
                    This probably happened because you are curating locally and
                    haven't added the right metadata to the curators.tsv file at
                    {CURATORS_PATH}.

                    You can fix this with the following steps:

                    1. Add a row to the curators.tsv file with your local machine's
                       username "{getpass.getuser()}" in the first column, your ORCID in
                       the second column, and your full name in the third column
                    2. Replace all instances of "{author.curie}" in {path}
                       with your ORCID, properly prefixed with `orcid:`
                    """).rstrip(),
                )

    def test_cross_redundancy(self) -> None:
        """Test the redundancy of manually curated mappings and predicted mappings."""
        counter: defaultdict[tuple[str, str, str, str], defaultdict[str, list[int]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for label, line, mapping in self._iter_groups():
            counter[get_canonical_tuple(mapping)][label].append(line)

        redundant = []
        for mapping_key, label_to_lines in counter.items():
            if len(label_to_lines) <= 1:
                continue
            redundant.append((mapping_key, sorted(label_to_lines.items())))

        if redundant:
            msg = "".join(
                f"\n  {mapping}: {_locations_str(locations)}" for mapping, locations in redundant
            )
            raise ValueError(f"{len(redundant)} are redundant: {msg}")

    def assert_no_internal_redundancies(self, mappings: list[SemanticMapping]) -> None:
        """Assert that the list of mappings doesn't have any redundancies."""
        counter: defaultdict[tuple[NamableReference, NamableReference], list[int]] = defaultdict(
            list
        )
        for line_number, mapping in enumerate(mappings, start=1):
            counter[mapping.subject, mapping.object].append(line_number)
        redundant = _extract_redundant(counter)
        if redundant:
            msg = "".join(
                f"\n  {subject.curie}/{obj.curie}: {locations}"
                for (subject, obj), locations in redundant
            )
            raise ValueError(f"{len(redundant)} are redundant: {msg}")

    def test_predictions_sorted(self) -> None:
        """Test the predictions are in a canonical order."""
        self.assertEqual(
            self.predictions,
            sorted(self.predictions, key=mapping_sort_key),
            msg="Predictions are not sorted",
        )
        self.assert_no_internal_redundancies(self.predictions)

    def test_curations_sorted(self) -> None:
        """Test the true curated mappings are in a canonical order."""
        self.assertEqual(
            self.mappings,
            sorted(self.mappings, key=mapping_sort_key),
            msg="True curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.mappings)

    def test_false_mappings_sorted(self) -> None:
        """Test the false curated mappings are in a canonical order."""
        self.assertEqual(
            self.incorrect,
            sorted(self.incorrect, key=mapping_sort_key),
            msg="False curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.incorrect)

    def test_unsure_sorted(self) -> None:
        """Test the unsure mappings are in a canonical order."""
        self.assertEqual(
            self.unsure,
            sorted(self.unsure, key=mapping_sort_key),
            msg="Unsure curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.unsure)


class PathIntegrityTestCase(IntegrityTestCase):
    """A test case that can be configured with paths.

    For example, in this might be used in a custom instance of Biomappings
    like in the following:

    .. code-block:: python

        from biomappings.testing import PathIntegrityTestCase

        HERE = Path(__file__).parent.resolve()


        class TestCustom(PathIntegrityTestCase):
            predictions_path = HERE.joinpath("predictions.tsv")
            positives_path = HERE.joinpath("positive.tsv")
            negatives_path = HERE.joinpath("negative.tsv")
            unsure_path = HERE.joinpath("unsure.tsv")
    """

    predictions_path: ClassVar[str | Path]
    positives_path: ClassVar[str | Path]
    negatives_path: ClassVar[str | Path]
    unsure_path: ClassVar[str | Path]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case."""
        cls.predictions = load_predictions(path=cls.predictions_path)
        cls.mappings = load_mappings(path=cls.positives_path)
        cls.incorrect = load_mappings(path=cls.negatives_path)
        cls.unsure = load_mappings(path=cls.unsure_path)



"""Update biomappings internal format to SSSOM."""

from __future__ import annotations

from pathlib import Path

import bioregistry
import pandas as pd

from biomappings.utils import (
    NEGATIVES_SSSOM_PATH,
    POSITIVES_SSSOM_PATH,
    PREDICTIONS_SSSOM_PATH,
    UNSURE_SSSOM_PATH,
)

HERE = Path(__file__).parent.resolve()
OLD_RESOURCE_FOLDER = Path.home().joinpath("Desktop", "biomappings-old")


def main() -> None:
    """Update biomappings internal format to SSSOM."""
    paths = [
        (OLD_RESOURCE_FOLDER.joinpath("incorrect.tsv"), NEGATIVES_SSSOM_PATH, True),
        (OLD_RESOURCE_FOLDER.joinpath("unsure.tsv"), UNSURE_SSSOM_PATH, False),
        (OLD_RESOURCE_FOLDER.joinpath("mappings.tsv"), POSITIVES_SSSOM_PATH, False),
    ]
    for old_path, new_path, add_not in paths:
        update_curated(old_path, new_path, add_not=add_not)

    update_predicted(OLD_RESOURCE_FOLDER.joinpath("predictions.tsv"), PREDICTIONS_SSSOM_PATH)


def update_predicted(old_path: Path, new_path: Path) -> None:
    """Run the update on a given predictions file."""
    df = pd.read_csv(old_path, sep="\t")
    df = _shared_update(df)
    df = df.rename(columns={"source": "mapping_tool"})
    df.to_csv(new_path, sep="\t", index=False)


def update_curated(old_path: Path, new_path: Path, add_not: bool = False) -> None:
    """Run the update on a given curated file."""
    df = pd.read_csv(old_path, sep="\t")
    df = _shared_update(df)

    # we're deleting prediction type and confidence since these
    # don't have a proper place in the SSSOM data model and can
    # be understood from the original script
    del df["prediction_type"]
    del df["prediction_confidence"]

    df = df.rename(
        columns={
            "source": "author_id",
            "prediction_source": "mapping_tool",
            "prediction_confidence": "confidence",
        }
    )

    if add_not:
        df["predicate_modifier"] = "Not"
    else:
        df["predicate_modifier"] = ""
    df.to_csv(new_path, sep="\t", index=False)


def _shared_update(df: pd.DataFrame) -> pd.DataFrame:
    df["source prefix"] = [
        bioregistry.NormalizedReference(prefix=p, identifier=i).curie
        for p, i in df[["source prefix", "source identifier"]].values
    ]
    df["target prefix"] = [
        bioregistry.NormalizedReference(prefix=p, identifier=i).curie
        for p, i in df[["target prefix", "target identifier"]].values
    ]
    del df["source identifier"]
    del df["target identifier"]

    df = df.rename(
        columns={
            "type": "mapping_justification",
            "source prefix": "subject_id",
            "target prefix": "object_id",
            "source name": "subject_label",
            "target name": "object_label",
            "relation": "predicate_id",
        }
    )
    return df


if __name__ == "__main__":
    main()



"""Community curated mappings between biomedical entities."""

from .graph import get_false_graph, get_predictions_graph, get_true_graph  # noqa:F401
from .resources import (  # noqa:F401
    SemanticMapping,
    load_false_mappings,
    load_mappings,
    load_mappings_subset,
    load_predictions,
    load_unsure,
)



"""Utilities for contributing back to upstream resources."""

from __future__ import annotations

from biomappings import SemanticMapping, load_mappings

__all__ = [
    "get_curated_mappings",
]


def get_curated_mappings(prefix: str) -> list[SemanticMapping]:
    """Get mappings for a given prefix."""
    mappings = []
    for mapping in load_mappings():
        if mapping.subject.prefix == prefix:
            mappings.append(mapping)
        elif mapping.object.prefix == prefix:
            mappings.append(mapping.flip())
    return mappings



"""Utilities."""

from __future__ import annotations

import os
from collections.abc import Mapping
from pathlib import Path
from subprocess import CalledProcessError, check_output

from bioregistry import NormalizedNamableReference

__all__ = [
    "RESOURCE_PATH",
    "CMapping",
    "get_canonical_tuple",
    "get_git_hash",
    "get_script_url",
]

HERE = Path(__file__).parent.resolve()
ROOT = HERE.parent.parent.resolve()
RESOURCE_PATH = HERE.joinpath("resources")

POSITIVES_SSSOM_PATH = RESOURCE_PATH.joinpath("positive.sssom.tsv")
NEGATIVES_SSSOM_PATH = RESOURCE_PATH.joinpath("negative.sssom.tsv")
UNSURE_SSSOM_PATH = RESOURCE_PATH.joinpath("unsure.sssom.tsv")
PREDICTIONS_SSSOM_PATH = RESOURCE_PATH.joinpath("predictions.sssom.tsv")
CURATORS_PATH = RESOURCE_PATH.joinpath("curators.tsv")

DOCS = ROOT.joinpath("docs")
IMG = DOCS.joinpath("img")
DATA = DOCS.joinpath("_data")


def get_git_hash() -> str | None:
    """Get the git hash.

    :return:
        The git hash, equals 'UNHASHED' if encountered CalledProcessError, signifying that the
        code is not installed in development mode.
    """
    rv = _git("rev-parse", "HEAD")
    if not rv:
        return None
    return rv[:6]


def commit(message: str) -> str | None:
    """Make a commit with the following message."""
    return _git("commit", "-m", message, "-a")


def push(branch_name: str | None = None) -> str | None:
    """Push the git repo."""
    if branch_name is not None:
        return _git("push", "origin", branch_name)
    else:
        return _git("push")


def not_main() -> bool:
    """Return if on the master branch."""
    return "master" != _git("rev-parse", "--abbrev-ref", "HEAD")


def get_branch() -> str:
    """Return current git branch."""
    rv = _git("branch", "--show-current")
    if rv is None:
        raise RuntimeError
    return rv


def _git(*args: str) -> str | None:
    with open(os.devnull, "w") as devnull:
        try:
            ret = check_output(  # noqa: S603
                ["git", *args],  # noqa:S607
                cwd=os.path.dirname(__file__),
                stderr=devnull,
            )
        except CalledProcessError as e:
            print(e)  # noqa:T201
            return None
        else:
            return ret.strip().decode("utf-8")


def get_script_url(fname: str) -> str:
    """Get the source path for this script.

    :param fname: Pass ``__file__`` as the argument to this function.
    :return: The script's URL to GitHub
    """
    commit_hash = get_git_hash()
    script_name = os.path.basename(fname)
    return f"https://github.com/biomappings/biomappings/blob/{commit_hash}/scripts/{script_name}"


def get_canonical_tuple(mapping) -> tuple[str, str, str, str]:
    """Get the canonical tuple from a mapping entry."""
    source = mapping.subject
    target = mapping.object
    if source > target:
        source, target = target, source
    return (*source.pair, *target.pair)


#: A filter 3-dictionary of source prefix to target prefix to source identifier to target identifier
CMapping = Mapping[str, Mapping[str, Mapping[str, str]]]

EXACT_MATCH = NormalizedNamableReference.from_curie("skos:exactMatch")
NARROW_MATCH = NormalizedNamableReference.from_curie("skos:narrowMatch")
BROAD_MATCH = NormalizedNamableReference.from_curie("skos:broadMatch")

LEXICAL_MATCHING_PROCESS = NormalizedNamableReference.from_curie("semapv:LexicalMatching")
MANUAL_MAPPING_CURATION = NormalizedNamableReference.from_curie("semapv:ManualMappingCuration")



"""Generate a summary for the Biomappings website."""

from __future__ import annotations

import itertools as itt
import os
import typing
from collections import Counter
from collections.abc import Iterable
from typing import Any, TypedDict, cast

import click
import yaml
from curies import NamableReference

from biomappings.resources import SemanticMapping

__all__ = [
    "export",
]


@click.command()
def export() -> None:
    """Create export data file."""
    from biomappings.resources import (
        load_false_mappings,
        load_mappings,
        load_predictions,
        load_unsure,
    )
    from biomappings.utils import DATA

    path = os.path.join(DATA, "summary.yml")

    true_mappings = load_mappings()
    false_mappings = load_false_mappings()
    unsure_mappings = load_unsure()

    rv: dict[str, Any] = {
        "contributors": _get_contributors(
            itt.chain(true_mappings, false_mappings, unsure_mappings)
        ),
    }

    for key, mappings in [
        ("positive", true_mappings),
        ("negative", false_mappings),
        ("unsure", unsure_mappings),
        ("predictions", load_predictions()),
    ]:
        count_records = _get_count_records(mappings)
        rv[key] = count_records
        rv[f"{key}_mapping_count"] = sum(count_record["count"] for count_record in count_records)
        rv[f"{key}_prefix_count"] = len(
            set(
                itt.chain.from_iterable(
                    (count_record["source"], count_record["target"])
                    for count_record in count_records
                )
            )
        )

    with open(path, "w") as file:
        yaml.safe_dump(rv, file, indent=2)


class CountRecord(TypedDict):
    """A result from getting counter."""

    source: str
    target: str
    count: int


def _get_count_records(mappings: Iterable[SemanticMapping]) -> list[CountRecord]:
    counter: typing.Counter[tuple[str, str]] = Counter()
    for mapping in mappings:
        subject, target = mapping.subject, mapping.object
        if subject > target:
            subject, target = target, subject
        counter[str(subject.prefix), str(target.prefix)] += 1
    return [
        {"source": subject_prefix, "target": target_prefix, "count": count}
        for (subject_prefix, target_prefix), count in counter.most_common()
    ]


def _get_contributors(mappings: Iterable[SemanticMapping]) -> list[dict[str, str | int]]:
    from biomappings.resources import load_curators

    orcid_to_curator_reference: dict[str, NamableReference] = {
        record.identifier: record for record in load_curators().values()
    }
    counter = Counter(
        mapping.author.identifier
        for mapping in mappings
        if mapping.author and mapping.author.prefix == "orcid"
    )
    return [
        {
            "count": count,
            "orcid": reference.identifier,
            "name": cast(str, reference.name),
        }
        if (reference := orcid_to_curator_reference.get(orcid)) is not None
        else {"count": count}
        for orcid, count in counter.most_common()
    ]


if __name__ == "__main__":
    export()



"""Export Biomappings as SSSOM."""

from __future__ import annotations

import importlib.metadata
import pathlib
from collections.abc import Collection
from typing import TYPE_CHECKING, NamedTuple

import bioregistry
import click
import yaml
from curies import ReferenceTuple
from tqdm.auto import tqdm

from biomappings.resources import NEGATIVES_SSSOM_PATH, POSITIVES_SSSOM_PATH, PREDICTIONS_SSSOM_PATH
from biomappings.utils import DATA

if TYPE_CHECKING:
    import pandas as pd
    from sssom import MappingSetDataFrame

DIRECTORY = pathlib.Path(DATA).joinpath("sssom")
DIRECTORY.mkdir(exist_ok=True, parents=True)
TSV_PATH = DIRECTORY.joinpath("biomappings.sssom.tsv")
JSON_PATH = DIRECTORY.joinpath("biomappings.sssom.json")
OWL_PATH = DIRECTORY.joinpath("biomappings.sssom.owl")
META_PATH = DIRECTORY.joinpath("biomappings.sssom.yml")

CC0_URL = "https://creativecommons.org/publicdomain/zero/1.0/"
META = {
    "license": CC0_URL,
    "mapping_provider": "https://github.com/biopragmatics/biomappings",
    "mapping_set_description": "Biomappings is a repository of community curated and predicted equivalences and "
    "related mappings between named biological entities that are not available from primary sources. It's also a "
    "place where anyone can contribute curations of predicted mappings or their own novel mappings.",
    "mapping_set_id": "https://w3id.org/biopragmatics/biomappings/sssom/biomappings.sssom.tsv",
    "mapping_set_title": "Biomappings",
    "mapping_set_version": importlib.metadata.version("biomappings"),
}

columns = [
    "subject_id",
    "subject_label",
    "predicate_id",
    "predicate_modifier",
    "object_id",
    "object_label",
    "mapping_justification",
    "author_id",
    "confidence",
    "mapping_tool",
]


class SSSOMReturnTuple(NamedTuple):
    """A package for SSSOM coalation."""

    prefix_map: dict[str, str]
    creator_curies: list[str]
    df: pd.DataFrame
    msdf: MappingSetDataFrame


def get_sssom_df(*, use_tqdm: bool = False) -> SSSOMReturnTuple:
    """Get an SSSOM dataframe."""
    import pandas as pd

    creator_curies: set[str] = set()
    prefixes: set[str] = {"semapv"}

    # NEW WAY: load all DFs, concat them, reorder columns
    a = pd.read_csv(POSITIVES_SSSOM_PATH, sep="\t")
    b = pd.read_csv(NEGATIVES_SSSOM_PATH, sep="\t")
    c = pd.read_csv(PREDICTIONS_SSSOM_PATH, sep="\t")
    df = pd.concat([a, b, c])
    df = df[columns]

    for column in ["subject_id", "object_id", "predicate_id"]:
        df[column] = df[column].map(lambda p: bioregistry.normalize_curie(p, use_preferred=True))

    for _, mapping in tqdm(
        df.iterrows(), desc="tabulating prefixes & authors", disable=not use_tqdm
    ):
        prefixes.add(_get_prefix(mapping["subject_id"]))
        prefixes.add(_get_prefix(mapping["predicate_id"]))
        prefixes.add(_get_prefix(mapping["object_id"]))
        author_id = mapping["author_id"]
        if pd.notna(author_id) and any(author_id.startswith(x) for x in ["orcid:", "wikidata:"]):
            prefixes.add(_get_prefix(author_id))
            creator_curies.add(author_id)
        # TODO add justification:

    from sssom.constants import DEFAULT_VALIDATION_TYPES
    from sssom.parsers import from_sssom_dataframe
    from sssom.validators import validate

    prefix_map = get_prefix_map(prefixes)

    try:
        msdf = from_sssom_dataframe(df, prefix_map=prefix_map, meta=META)
    except Exception as e:
        click.secho(f"SSSOM Export failed...\n{e}", fg="red")
        raise

    results = validate(msdf=msdf, validation_types=DEFAULT_VALIDATION_TYPES, fail_on_error=False)
    for validator_type, validation_report in results.items():
        if validation_report.results:
            click.secho(f"SSSOM Validator Failed: {validator_type}", fg="red")
            for result in validation_report.results:
                click.secho(f"- {result}", fg="red")
            click.echo("")

    return SSSOMReturnTuple(prefix_map, sorted(creator_curies), df, msdf)


def _get_prefix(curie: str) -> str:
    """Get a prefix from a CURIE string."""
    return ReferenceTuple.from_curie(curie).prefix


def get_prefix_map(prefixes: Collection[str]) -> dict[str, str]:
    """Get a CURIE map containing only the relevant prefixes."""
    prefix_map = {}
    for prefix in sorted(prefixes, key=str.casefold):
        resource = bioregistry.get_resource(prefix)
        if resource is None:
            raise KeyError
        uri_prefix = resource.get_rdf_uri_prefix() or resource.get_uri_prefix()
        if uri_prefix is None:
            raise ValueError(f"could not look up URI prefix for {prefix}")
        preferred_prefix = resource.get_preferred_prefix() or prefix
        prefix_map[preferred_prefix] = uri_prefix
    return prefix_map


@click.command()
def sssom() -> None:
    """Export SSSOM."""
    from sssom.writers import write_json, write_owl

    prefix_map, creator_curies, df, msdf = get_sssom_df()

    tsv_meta = {**META, "creator_id": creator_curies, "curie_map": prefix_map}

    with TSV_PATH.open("w") as file:
        for line in yaml.safe_dump(tsv_meta).splitlines():
            print(f"# {line}", file=file)
        df.to_csv(file, sep="\t", index=False)

    with open(META_PATH, "w") as file:
        yaml.safe_dump(tsv_meta, file)

    click.echo("Writing JSON")
    with JSON_PATH.open("w") as file:
        msdf.metadata["mapping_set_id"] = (
            "https://w3id.org/biopragmatics/biomappings/sssom/biomappings.sssom.json"
        )
        write_json(msdf, file)
    click.echo("Writing OWL")
    with OWL_PATH.open("w") as file:
        msdf.metadata["mapping_set_id"] = (
            "https://w3id.org/biopragmatics/biomappings/sssom/biomappings.sssom.owl"
        )
        write_owl(msdf, file)


if __name__ == "__main__":
    sssom()



"""The biomappings CLI."""

from __future__ import annotations

import sys
from pathlib import Path

import click
from more_click import run_app

from .graph import charts
from .resources.export_sssom import sssom
from .summary import export
from .upload_ndex import ndex
from .utils import get_git_hash


@click.group()
@click.version_option()
def main() -> None:
    """Run the biomappings CLI."""


if get_git_hash() is not None:

    @main.command()
    @click.option("--predictions-path", type=click.Path(), help="A predictions TSV file path")
    @click.option("--positives-path", type=click.Path(), help="A positives curation TSV file path")
    @click.option("--negatives-path", type=click.Path(), help="A negatives curation TSV file path")
    @click.option("--unsure-path", type=click.Path(), help="An unsure curation TSV file path")
    def web(
        predictions_path: Path,
        positives_path: Path,
        negatives_path: Path,
        unsure_path: Path,
    ) -> None:
        """Run the biomappings web app."""
        from .wsgi import get_app

        app = get_app(
            predictions_path=predictions_path,
            positives_path=positives_path,
            negatives_path=negatives_path,
            unsure_path=unsure_path,
        )
        run_app(app, with_gunicorn=False)

    @main.command()
    @click.option("--path", required=True, type=click.Path(), help="A predictions TSV file path")
    def curate(path: Path) -> None:
        """Run a target curation web app."""
        from curies import NamableReference

        from .resources import _load_table
        from .wsgi import get_app

        target_references: list[NamableReference] = []
        for mapping in _load_table(path, standardize=True):
            target_references.append(mapping.subject)
            target_references.append(mapping.object)
        app = get_app(target_references=target_references)
        run_app(app, with_gunicorn=False)

else:

    @main.command()
    def web() -> None:
        """Show an error for the web interface."""
        click.secho(
            "You are not running biomappings from a development installation.\n"
            "Please run the following to install in development mode:\n"
            "  $ git clone https://github.com/biomappings/biomappings.git\n"
            "  $ cd biomappings\n"
            "  $ pip install -e .",
            fg="red",
        )


@main.command()
@click.pass_context
def update(ctx: click.Context) -> None:
    """Run all update functions."""
    click.secho("Building general exports", fg="green")
    ctx.invoke(export)
    click.secho("Building SSSOM export", fg="green")
    ctx.invoke(sssom)
    click.secho("Generating charts", fg="green")
    ctx.invoke(charts)


@main.command()
def lint() -> None:
    """Sort files and remove duplicates."""
    from . import resources

    resources.lint_true_mappings(standardize=True)
    resources.lint_false_mappings(standardize=True)
    resources.lint_unsure_mappings(standardize=True)
    resources.lint_predictions(standardize=True)


@main.command()
@click.argument("prefixes", nargs=-1)
def prune(prefixes: list[str]) -> None:
    """Prune inferred mappings between the given prefixes from the predictions."""
    from .mapping_graph import get_mutual_mapping_filter
    from .resources import filter_predictions

    if len(prefixes) < 2:
        click.secho("Must give at least 2 prefixes", fg="red")
        sys.exit(0)

    cf = get_mutual_mapping_filter(prefixes[0], prefixes[1:])
    filter_predictions(cf)


@main.command()
def remove_curated() -> None:
    """Remove curated mappings from the predicted mappings, use this if they get out of sync."""
    from .resources import filter_predictions, get_curated_filter

    filter_predictions(get_curated_filter())


main.add_command(export)
main.add_command(ndex)
main.add_command(charts)
main.add_command(sssom)

if __name__ == "__main__":
    main()



"""Contribute Biomappings back to ontologies encoded in the OBO flat file format.

Example ontologies using the OBO flat file format:
- Uber Anatomy Ontology (UBERON)
- Mondo Disease Ontology (MONDO)
"""

from __future__ import annotations

from copy import deepcopy
from pathlib import Path

import bioregistry
import click
from tqdm.auto import tqdm

from biomappings import SemanticMapping
from biomappings.contribute.utils import get_curated_mappings


def update_obo(*, prefix: str, path: str | Path) -> None:
    """Update an OBO flat file.

    :param prefix: Prefix for the ontology
    :param path: Path to the ontology edit file, encoded with OBO flat file format

    Example usage

    .. code-block:: sh

        git clone https://github.com/obophenotype/uberon.git
        python -m biomappings.contribute.obo --prefix uberon --path uberon/src/ontology/uberon-edit.obo
    """
    path = Path(path).resolve()
    with path.open("r") as file:
        lines = file.readlines()
    mappings = get_curated_mappings(prefix)
    lines = update_obo_lines(lines=lines, mappings=mappings)
    with path.open("w") as file:
        file.writelines(lines)


def update_obo_lines(
    *, lines: list[str], mappings: list[SemanticMapping], progress: bool = True
) -> list[str]:
    """Update the lines of an OBO file.

    :param mappings: Mappings to add
    :param lines: A list of lines of the file (still containing trailing newlines)
    :param progress: Show a progress bar
    :returns: New lines. Does not modify the original list.
    """
    lines = deepcopy(lines)

    for mapping in tqdm(mappings, unit="mapping", unit_scale=True, disable=not progress):
        subject_curie = bioregistry.normalize_curie(
            mapping.subject.curie, use_preferred=True, strict=True
        )
        if subject_curie is None:
            raise ValueError
        object_curie = bioregistry.normalize_curie(
            mapping.object.curie, use_preferred=True, strict=True
        )
        if object_curie is None:
            raise ValueError

        if not mapping.author or mapping.author.prefix != "orcid":
            continue

        lines = add_xref(
            lines,
            node_curie=subject_curie,
            xref_curie=object_curie,
            xref_name=mapping.object.name,
            author_orcid=mapping.author.identifier,
        )
    return lines


def add_xref(
    lines: list[str], *, node_curie: str, xref_curie: str, xref_name: str, author_orcid: str
) -> list[str]:
    """Add xref to OBO file lines in the appropriate place."""
    look_for_xref = False
    id_idx = None

    #: The 0-indexed line number on which the first xref appears
    start_xref_idx = None
    #: The 0-indexed line number on which the definition appears in a given
    def_idx = None
    xref_entries = []
    xref_values = set()
    for idx, line in enumerate(lines):
        if line == f"id: {node_curie}":
            id_idx = idx
            look_for_xref = True
        if look_for_xref and line.startswith("def"):
            def_idx = idx
        if look_for_xref:
            if line.startswith("xref"):
                if not start_xref_idx:
                    start_xref_idx = idx
                xref_line: str = line[len("xref:") :].strip()
                xref_values.add(_extract_ref(xref_line))
                xref_entries.append(xref_line)
            if start_xref_idx and not line.startswith("xref"):
                break
        if look_for_xref and not line.strip():
            start_xref_idx = def_idx

    if id_idx is None:
        # term not found
        return lines

    if xref_curie in xref_values:
        # term already has this xref, don't modify it
        return lines

    if start_xref_idx is None:
        if def_idx:
            start_xref_idx = def_idx + 1
        else:
            # there were no existing xrefs, so let's just stick them directly after the definition
            start_xref_idx = id_idx + 1

    xref_entries.append(xref_curie)
    xref_entries = sorted(xref_entries)
    xr_idx = xref_entries.index(xref_curie)
    line = f'xref: {xref_curie} {{dcterms:contributor="https://orcid.org/{author_orcid}"}} ! {xref_name}'
    lines.insert(start_xref_idx + xr_idx, line)
    return lines


def _extract_ref(xref_line: str) -> str:
    # remove trailing comments
    if "!" in xref_line:
        xref_line = xref_line[: xref_line.find("!")].strip()
    # remove meta-xrefs
    if "[" in xref_line:
        xref_line = xref_line[: xref_line.find("[")].strip()
    # if there are qualifiers, only keep everything up until them
    if "{" in xref_line:
        return xref_line[: xref_line.find("{")].strip()
    return xref_line


@click.command()
@click.option("--prefix", required=True, help="The prefix corresponding to the ontology")
@click.option(
    "--path",
    type=click.Path(),
    required=True,
    help="After forking and cloning the version controlled repository for an ontology locally, "
    "give the path inside the directory to the ontology's edit file.",
)
def main(prefix: str, path: Path) -> None:
    """Contribute to an OBO file."""
    update_obo(prefix=prefix, path=path)


if __name__ == "__main__":
    main()



"""Load Biomappings as a graph."""

from __future__ import annotations

import itertools as itt
import logging
import os
from collections import Counter
from collections.abc import Collection, Iterable, Sequence
from typing import TYPE_CHECKING

import click
import networkx as nx
import yaml
from curies import Reference
from tqdm import tqdm

from biomappings.resources import (
    SemanticMapping,
    load_false_mappings,
    load_mappings,
    load_predictions,
)
from biomappings.utils import DATA, EXACT_MATCH, IMG

if TYPE_CHECKING:
    import matplotlib.axes

__all__ = [
    "get_false_graph",
    "get_predictions_graph",
    "get_true_graph",
]

logger = logging.getLogger(__name__)


def get_true_graph(
    include: Sequence[Reference] | None = None,
    exclude: Sequence[Reference] | None = None,
) -> nx.Graph:
    """Get a graph of the true mappings."""
    return _graph_from_mappings(load_mappings(), strata="correct", include=include, exclude=exclude)


def get_false_graph(
    include: Sequence[Reference] | None = None,
    exclude: Sequence[Reference] | None = None,
) -> nx.Graph:
    """Get a graph of the false mappings."""
    return _graph_from_mappings(
        load_false_mappings(), strata="incorrect", include=include, exclude=exclude
    )


def get_predictions_graph(
    include: Collection[Reference] | None = None,
    exclude: Collection[Reference] | None = None,
) -> nx.Graph:
    """Get a graph of the predicted mappings."""
    return _graph_from_mappings(
        load_predictions(), strata="predicted", include=include, exclude=exclude
    )


def _graph_from_mappings(
    mappings: Iterable[SemanticMapping],
    strata: str,
    include: Collection[Reference] | None = None,
    exclude: Collection[Reference] | None = None,
) -> nx.Graph:
    graph = nx.Graph()

    if include is not None:
        include = set(include)
        logger.info("only including %s", include)
    if exclude is not None:
        exclude = set(exclude)
        logger.info("excluding %s", exclude)

    for mapping in mappings:
        if exclude and (mapping.predicate in exclude):
            continue
        if include and (mapping.predicate not in include):
            continue
        graph.add_edge(
            mapping.subject,
            mapping.object,
            relation=mapping.predicate.curie,
            provenance=mapping.author.name if mapping.author else None,
            type=mapping.mapping_justification.curie,
            strata=strata,
        )
    return graph


@click.command()
def charts() -> None:
    """Make charts."""
    import matplotlib.pyplot as plt
    import seaborn as sns

    true_mappings = load_mappings()
    true_graph = _graph_from_mappings(true_mappings, include=[EXACT_MATCH], strata="correct")
    for u, v in true_graph.edges():
        true_graph.edges[u, v]["correct"] = True
    false_mappings = load_false_mappings()
    false_graph = _graph_from_mappings(false_mappings, include=[EXACT_MATCH], strata="incorrect")
    for u, v in false_graph.edges():
        false_graph.edges[u, v]["correct"] = False

    component_node_sizes, component_edge_sizes, component_densities, component_number_prefixes = (
        [],
        [],
        [],
        [],
    )
    prefix_list = []
    components_with_duplicate_prefixes = []
    incomplete_components = []
    unstable_components = []
    n_duplicates = []
    for component in tqdm(
        nx.connected_components(true_graph), desc="Positive SCCs", unit="component", unit_scale=True
    ):
        component = true_graph.subgraph(component)
        node_size = component.number_of_nodes()
        edge_size = component.number_of_edges()

        nodes_data = {
            reference.curie: {
                "link": f"https://bioregistry.io/{reference.curie}",
                "prefix": str(reference.prefix),
                "identifier": reference.identifier,
                "name": reference.name,
            }
            for reference in component.nodes
        }

        unstable_edges = [
            (u, v, false_graph[u, v])
            for u, v in itt.combinations(component, 2)
            if false_graph.has_edge(u, v)
        ]
        if unstable_edges:
            unstable_components.append({"nodes": nodes_data, "unstable_edges": unstable_edges})

        component_node_sizes.append(node_size)
        component_edge_sizes.append(edge_size)
        if node_size > 2:
            component_densities.append(nx.density(component))
        if node_size > 2 and edge_size < (node_size * (node_size - 1) / 2):
            incomplete_components_edges = []
            for u, v in sorted(nx.complement(component.copy()).edges()):
                if u > v:
                    u, v = v, u
                incomplete_components_edges.append(
                    {
                        "source": {"curie": u.curie, **nodes_data[u.curie]},
                        "target": {"curie": v.curie, **nodes_data[v.curie]},
                    }
                )
            incomplete_components_edges = sorted(
                incomplete_components_edges, key=lambda d: d["source"]["curie"]
            )
            incomplete_components.append(
                {
                    "nodes": nodes_data,
                    "edges": incomplete_components_edges,
                }
            )

        prefixes = [node.prefix for node in component]
        prefix_list.extend(prefixes)
        unique_prefixes = len(set(prefixes))
        component_number_prefixes.append(unique_prefixes)
        _n_duplicates = len(prefixes) - unique_prefixes
        n_duplicates.append(_n_duplicates)
        if _n_duplicates:
            components_with_duplicate_prefixes.append(nodes_data)

    with open(os.path.join(DATA, "incomplete_components.yml"), "w") as file:
        yaml.safe_dump(incomplete_components, file)
    with open(os.path.join(DATA, "components_with_duplicate_prefixes.yml"), "w") as file:
        yaml.safe_dump(components_with_duplicate_prefixes, file)
    with open(os.path.join(DATA, "unstable_components.yml"), "w") as file:
        yaml.safe_dump(unstable_components, file)

    fig, axes = plt.subplots(2, 3, figsize=(10.5, 6.5))

    _countplot_list(component_node_sizes, ax=axes[0][0])
    axes[0][0].set_yscale("log")
    axes[0][0].set_title("Size (Nodes)")

    _countplot_list(component_edge_sizes, ax=axes[0][1])
    axes[0][1].set_yscale("log")
    axes[0][1].set_title("Size (Edges)")
    axes[0][1].set_ylabel("")

    sns.kdeplot(component_densities, ax=axes[0][2])
    axes[0][2].set_xlim([0.0, 1.0])
    axes[0][2].set_title("Density ($|V| > 2$)")
    axes[0][2].set_ylabel("")

    _countplot_list(component_number_prefixes, ax=axes[1][0])
    axes[1][0].set_title("Number Prefixes")
    axes[1][0].set_yscale("log")
    # has duplicate prefix in component

    _countplot_list(n_duplicates, ax=axes[1][1])
    axes[1][1].set_yscale("log")
    axes[0][2].set_ylabel("")
    axes[1][1].set_title("Number Duplicate Prefixes")

    axes[1][2].axis("off")

    path = os.path.join(IMG, "components.png")
    click.echo(f"saving to {path}")
    plt.tight_layout()
    plt.savefig(path, dpi=300)
    plt.savefig(os.path.join(IMG, "components.svg"))
    plt.close(fig)

    prefix_counter = Counter(prefix_list)
    fig, axes = plt.subplots(1, 2, figsize=(8.5, 7))
    sns.countplot(y=prefix_list, ax=axes[0], order=[k for k, _ in prefix_counter.most_common()])
    axes[0].set_xscale("log")
    axes[0].set_xlabel("Count")
    axes[0].set_title(f"Prefixes ({len(prefix_counter)})")

    relations = [m.predicate.curie for m in true_mappings]
    relation_counter = Counter(relations)
    sns.countplot(y=relations, ax=axes[1], order=[k for k, _ in relation_counter.most_common()])
    axes[1].set_xscale("log")
    axes[1].set_xlabel("Count")
    axes[1].set_title(f"Relations ({len(relation_counter)})")

    path = os.path.join(IMG, "summary.png")
    click.echo(f"saving to {path}")
    plt.tight_layout()
    plt.savefig(path, dpi=300)
    plt.savefig(os.path.join(IMG, "summary.svg"))
    plt.close(fig)


def _countplot_list(data: list[int], ax: matplotlib.axes.Axes) -> None:
    import pandas as pd
    import seaborn as sns

    counter = Counter(data)
    for size in range(min(counter), max(counter)):
        if size not in counter:
            counter[size] = 0
    df = pd.DataFrame(counter.items(), columns=["size", "count"]).sort_values("size").reset_index()
    sns.barplot(data=df, x="size", y="count", ax=ax)


if __name__ == "__main__":
    charts()



"""Biomappings resources."""

from __future__ import annotations

import csv
import itertools as itt
import logging
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Callable, NamedTuple

from bioregistry import NormalizedNamableReference, NormalizedNamedReference
from curies import NamableReference, NamedReference
from pydantic import BaseModel, ConfigDict, Field
from tqdm.auto import tqdm
from typing_extensions import Literal, Self

from biomappings.utils import (
    CURATORS_PATH,
    NEGATIVES_SSSOM_PATH,
    POSITIVES_SSSOM_PATH,
    PREDICTIONS_SSSOM_PATH,
    UNSURE_SSSOM_PATH,
    get_canonical_tuple,
)

if TYPE_CHECKING:
    import semra

__all__ = [
    "SemanticMapping",
    "append_false_mappings",
    "append_prediction_tuples",
    "append_predictions",
    "append_true_mapping_tuples",
    "append_true_mappings",
    "append_unsure_mappings",
    "filter_predictions",
    "get_curated_filter",
    "load_curators",
    "load_false_mappings",
    "load_mappings",
    "load_mappings_subset",
    "load_predictions",
    "load_unsure",
    "prediction_tuples_from_semra",
    "remove_mappings",
    "write_false_mappings",
    "write_predictions",
    "write_true_mappings",
    "write_unsure_mappings",
]

logger = logging.getLogger(__name__)


class _CuratedTuple(NamedTuple):
    """A tuple for writing manual curations to SSSOM TSV."""

    subject_id: str
    subject_label: str
    predicate_id: str
    object_id: str
    object_label: str
    mapping_justification: str
    author_id: str
    mapping_tool: str
    predicate_modifier: str


class _PredictedTuple(NamedTuple):
    """A tuple for writing predictions to SSSOM TSV."""

    subject_id: str
    subject_label: str
    predicate_id: str
    object_id: str
    object_label: str
    mapping_justification: str
    confidence: str
    mapping_tool: str


class SemanticMapping(BaseModel):
    """A class for semantic mappings."""

    model_config = ConfigDict(frozen=True)

    subject: NamableReference
    predicate: NamableReference
    object: NamableReference
    mapping_justification: NamableReference = Field(
        ...,
        description="""\
        A `semapv <https://bioregistry.io/registry/semapv>`_ term describing the mapping type.

        These are relatively high level, and can be any child of ``semapv:Matching``, including:

        1. ``semapv:LexicalMatching``
        2. ``semapv:LogicalReasoning``
        """,
    )
    author: NamableReference | None = None
    mapping_tool: str | None = Field(
        None,
        description="""\
            The script or process that generated this mapping.

            Most of these scripts are in https://github.com/biopragmatics/biomappings/tree/master/scripts,
            or can be based off of them.
        """,
    )
    predicate_modifier: Literal["Not"] | None = None
    confidence: float | None = Field(
        None,
        description="""\
    An assessment of the confidence of the mapping, reported by the method used to generate it.

    This means that confidence values aren't generally comparable, though they should follow
    the rough standard that closer to 1 is more confident and closer to 0 is less confident.

    Most of the lexical mappings already in Biomappings were generated with Gilda.
    Depending on the script, the score therefore refers to either:

    1. The Gilda match score, inspired by https://aclanthology.org/W15-3801/. Section 5.2 of the
       `supplementary material for the Gilda paper <https://doi.org/10.1093/bioadv/vbac034>`_
       describes this score in detail, where 1.0 is best and 0 is worst.
       https://github.com/biopragmatics/biomappings/blob/master/scripts/generate_agrovoc_mappings.py
       is an example that uses this variant.
    2. A high-level estimation of the precision of the scores generated by the given script.
       For example, the CL-MeSH mappings were estimated to be 90% correct, so all the mappings
       generated by https://github.com/biopragmatics/biomappings/blob/master/scripts/generate_cl_mesh_mappings.py
       are marked with 0.9 as its score.

    However, other variants are possible. For example, this confidence could reflect the loss function
    if a knowledge graph embedding model was used ot generate a mapping prediction.
    """,
    )

    def flip(self) -> Self:
        """Flip the mapping, if it's an exact match."""
        if self.predicate.curie != "skos:exactMatch":
            raise NotImplementedError
        return self.model_copy(
            update={
                "subject": self.object,
                "object": self.subject,
            }
        )

    def as_curated_row(self) -> _CuratedTuple:
        """Get a row for a curated SSSOM TSV."""
        return _CuratedTuple(
            self.subject.curie,
            self.subject.name or "",
            self.predicate.curie,
            self.object.curie,
            self.object.name or "",
            self.mapping_justification.curie,
            self.author.curie if self.author is not None else "",
            self.mapping_tool or "",
            str(self.predicate_modifier or ""),
        )

    @classmethod
    def from_row(cls, row: dict[str, str], reference_cls: type[NamableReference]) -> Self:
        """Construct from a dictionary."""
        for k in ["subject", "object", "predicate", "author"]:
            k1 = f"{k}_id"
            v1 = row.pop(k1, None)
            if v1 is None:
                continue
            if not isinstance(v1, str):
                raise TypeError
            k2 = f"{k}_label"
            # type ignore because properly typing the dict i/o is a pain
            row[k] = reference_cls.from_curie(v1, name=row.pop(k2, None))  # type:ignore[assignment]
        for k in ["mapping_justification"]:
            row[k] = reference_cls.from_curie(row.pop(k))  # type:ignore[assignment]
        return cls.model_validate(row)

    def as_predicted_row(self) -> _PredictedTuple:
        """Get a row for a predicted SSSOM TSV."""
        return _PredictedTuple(
            self.subject.curie,
            self.subject.name or "",
            self.predicate.curie,
            self.object.curie,
            self.object.name or "",
            self.mapping_justification.curie,
            str(self.confidence) if self.confidence is not None else "",
            self.mapping_tool or "",
        )


def _load_table(path: str | Path, *, standardize: bool) -> list[SemanticMapping]:
    reference_cls: type[NamableReference]
    if standardize:
        reference_cls = NormalizedNamableReference
    else:
        reference_cls = NamableReference

    with Path(path).expanduser().resolve().open() as file:
        return [
            SemanticMapping.from_row(
                {
                    key: value_stripped
                    for key, value in record.items()
                    if value and (value_stripped := value.strip()) and value_stripped != "."
                },
                reference_cls,
            )
            for record in csv.DictReader(file, delimiter="\t")
        ]


def _write_helper(
    mappings: Iterable[SemanticMapping],
    path: str | Path,
    mode: Literal["w", "a"],
    t: Literal["curated", "predicted"],
) -> None:
    mappings = sorted(set(mappings), key=mapping_sort_key)
    header: Sequence[str]
    to_row: Callable[[SemanticMapping], _PredictedTuple | _CuratedTuple]
    if t == "curated":
        header = _CuratedTuple._fields
        to_row = SemanticMapping.as_curated_row
    else:
        header = _PredictedTuple._fields
        to_row = SemanticMapping.as_predicted_row
    path = Path(path).expanduser().resolve()
    with path.open(mode) as file:
        if mode == "w":
            print(*header, sep="\t", file=file)
        for mapping in mappings:
            print(*to_row(mapping), sep="\t", file=file)


def mapping_sort_key(mapping: SemanticMapping) -> tuple[str, ...]:
    """Return a tuple for sorting mapping dictionaries."""
    return (
        mapping.subject.curie,
        mapping.predicate.curie,
        mapping.object.curie,
        mapping.mapping_justification.curie,
        mapping.mapping_tool or "",
    )


def load_mappings(
    *, path: str | Path | None = None, standardize: bool = False
) -> list[SemanticMapping]:
    """Load the mappings table."""
    return _load_table(path or POSITIVES_SSSOM_PATH, standardize=standardize)


def load_mappings_subset(source: str, target: str) -> Mapping[str, str]:
    """Get a dictionary of 1-1 mappings from the source prefix to the target prefix."""
    # TODO replace with SeMRA functionality?
    return {
        mapping.subject.identifier: mapping.object.identifier
        for mapping in load_mappings()
        if mapping.subject.prefix == source and mapping.object.prefix == target
    }


def append_true_mappings(
    mappings: Iterable[SemanticMapping],
    *,
    sort: bool = True,
    path: Path | None = None,
    standardize: bool = False,
) -> None:
    """Append new lines to the mappings table."""
    if path is None:
        path = POSITIVES_SSSOM_PATH
    _write_helper(mappings, path=path, mode="a", t="curated")
    if sort:
        lint_true_mappings(path=path, standardize=standardize)


def append_true_mapping_tuples(mappings: Iterable[SemanticMapping]) -> None:
    """Append new lines to the mappings table."""
    append_true_mappings(mappings)


def write_true_mappings(mappings: Iterable[SemanticMapping], *, path: Path | None = None) -> None:
    """Write mappigns to the true mappings file."""
    _write_helper(mappings, path=path or POSITIVES_SSSOM_PATH, mode="w", t="curated")


def lint_true_mappings(*, path: Path | None = None, standardize: bool) -> None:
    """Lint the true mappings file."""
    _lint_curated_mappings(path=path or POSITIVES_SSSOM_PATH, standardize=standardize)


def _lint_curated_mappings(path: Path, *, standardize: bool) -> None:
    """Lint the true mappings file."""
    mapping_list = _load_table(path, standardize=standardize)
    mappings = _remove_redundant(mapping_list)
    _write_helper(mappings, path, mode="w", t="curated")


def load_false_mappings(
    *, path: Path | None = None, standardize: bool = False
) -> list[SemanticMapping]:
    """Load the false mappings table."""
    return _load_table(path or NEGATIVES_SSSOM_PATH, standardize=standardize)


def append_false_mappings(
    mappings: Iterable[SemanticMapping],
    *,
    sort: bool = True,
    path: Path | None = None,
    standardize: bool = False,
) -> None:
    """Append new lines to the false mappings table."""
    if path is None:
        path = NEGATIVES_SSSOM_PATH
    _write_helper(mappings=mappings, path=path, mode="a", t="curated")
    if sort:
        lint_false_mappings(path=path, standardize=standardize)


def write_false_mappings(mappings: Iterable[SemanticMapping], *, path: Path | None = None) -> None:
    """Write mappings to the false mappings file."""
    _write_helper(mappings, path or NEGATIVES_SSSOM_PATH, mode="w", t="curated")


def lint_false_mappings(*, path: Path | None = None, standardize: bool) -> None:
    """Lint the false mappings file."""
    _lint_curated_mappings(path=path or NEGATIVES_SSSOM_PATH, standardize=standardize)


def load_unsure(*, path: Path | None = None, standardize: bool = False) -> list[SemanticMapping]:
    """Load the unsure table."""
    return _load_table(path or UNSURE_SSSOM_PATH, standardize=standardize)


def append_unsure_mappings(
    mappings: Iterable[SemanticMapping],
    *,
    sort: bool = True,
    path: Path | None = None,
    standardize: bool = False,
) -> None:
    """Append new lines to the "unsure" mappings table."""
    if path is None:
        path = UNSURE_SSSOM_PATH
    _write_helper(mappings, path=path, mode="a", t="curated")
    if sort:
        lint_unsure_mappings(path=path, standardize=standardize)


def write_unsure_mappings(mappings: Iterable[SemanticMapping], *, path: Path | None = None) -> None:
    """Write mappings to the unsure mappings file."""
    _write_helper(mappings, path or UNSURE_SSSOM_PATH, mode="w", t="curated")


def lint_unsure_mappings(*, standardize: bool, path: Path | None = None) -> None:
    """Lint the unsure mappings file."""
    _lint_curated_mappings(path=path or UNSURE_SSSOM_PATH, standardize=standardize)


def load_predictions(
    *, path: str | Path | None = None, standardize: bool = False
) -> list[SemanticMapping]:
    """Load the predictions table."""
    return _load_table(path or PREDICTIONS_SSSOM_PATH, standardize=standardize)


def write_predictions(mappings: Iterable[SemanticMapping], *, path: Path | None = None) -> None:
    """Write new content to the predictions table."""
    _write_helper(mappings, path or PREDICTIONS_SSSOM_PATH, mode="w", t="predicted")


def append_prediction_tuples(
    prediction_tuples: Iterable[SemanticMapping],
    *,
    deduplicate: bool = True,
    sort: bool = True,
    path: Path | None = None,
    standardize: bool = False,
) -> None:
    """Append new lines to the predictions table that come as canonical tuples."""
    append_predictions(
        prediction_tuples, deduplicate=deduplicate, sort=sort, path=path, standardize=standardize
    )


def append_predictions(
    mappings: Iterable[SemanticMapping],
    *,
    deduplicate: bool = True,
    sort: bool = True,
    path: Path | None = None,
    standardize: bool,
) -> None:
    """Append new lines to the predictions table."""
    if deduplicate:
        existing_mappings = {
            get_canonical_tuple(existing_mapping)
            for existing_mapping in itt.chain(
                load_mappings(),
                load_false_mappings(),
                load_unsure(),
                load_predictions(),
            )
        }
        mappings = (
            mapping for mapping in mappings if get_canonical_tuple(mapping) not in existing_mappings
        )

    if path is None:
        path = PREDICTIONS_SSSOM_PATH
    _write_helper(mappings, path, mode="a", t="predicted")
    if sort:
        lint_predictions(path=path, standardize=standardize)


def lint_predictions(
    *,
    path: Path | None = None,
    additional_curated_mappings: Iterable[SemanticMapping] | None = None,
    standardize: bool,
) -> None:
    """Lint the predictions file.

    1. Make sure there are no redundant rows
    2. Make sure no rows in predictions match a row in the curated files
    3. Make sure it's sorted

    :param path: The path to the predicted mappings
    :param additional_curated_mappings: A list of additional mappings
    """
    mappings = remove_mappings(
        load_predictions(path=path, standardize=standardize),
        itt.chain(
            load_mappings(standardize=standardize),
            load_false_mappings(standardize=standardize),
            load_unsure(standardize=standardize),
            additional_curated_mappings or [],
        ),
    )
    mappings = _remove_redundant(mappings)
    mappings = sorted(mappings, key=mapping_sort_key)
    write_predictions(mappings, path=path)


def remove_mappings(
    mappings: Iterable[SemanticMapping], mappings_to_remove: Iterable[SemanticMapping]
) -> Iterable[SemanticMapping]:
    """Remove the first set of mappings from the second."""
    skip_tuples = {get_canonical_tuple(mtr) for mtr in mappings_to_remove}
    return (mapping for mapping in mappings if get_canonical_tuple(mapping) not in skip_tuples)


def _remove_redundant(mappings: Iterable[SemanticMapping]) -> Iterable[SemanticMapping]:
    dd = defaultdict(list)
    for mapping in mappings:
        dd[get_canonical_tuple(mapping)].append(mapping)
    return (max(mappings, key=_pick_best) for mappings in dd.values())


def _pick_best(mapping: SemanticMapping) -> int:
    """Assign a value for this mapping.

    :param mapping: A mapping dictionary
    :returns: An integer, where higher means a better choice.

    This function is currently simple, but can later be extended to
    account for several other things including:

    - confidence in the curator
    - prediction methodology
    - date of prediction/curation (to keep the earliest)
    """
    if mapping.author and mapping.author.prefix == "orcid":
        return 1
    return 0


def load_curators() -> dict[str, NamedReference]:
    """Load the curators table."""
    with CURATORS_PATH.open() as file:
        return {
            record["user"]: NormalizedNamedReference(
                prefix="orcid", identifier=record["orcid"], name=record["name"]
            )
            for record in csv.DictReader(file, delimiter="\t")
        }


def filter_predictions(custom_filter: Mapping[str, Mapping[str, Mapping[str, str]]]) -> None:
    """Filter all the predictions by removing what's in the custom filter then re-write.

    :param custom_filter: A filter 3-dictionary of source prefix to target prefix
        to source identifier to target identifier
    """
    predictions = load_predictions()
    predictions = [
        prediction for prediction in predictions if _check_filter(prediction, custom_filter)
    ]
    write_predictions(predictions)


def _check_filter(
    prediction: SemanticMapping,
    custom_filter: Mapping[str, Mapping[str, Mapping[str, str]]],
) -> bool:
    v = (
        custom_filter.get(prediction.subject.prefix, {})
        .get(prediction.object.prefix, {})
        .get(prediction.subject.identifier)
    )
    return prediction.object.identifier != v


def get_curated_filter() -> Mapping[str, Mapping[str, Mapping[str, str]]]:
    """Get a filter over all curated mappings."""
    d: defaultdict[str, defaultdict[str, dict[str, str]]] = defaultdict(lambda: defaultdict(dict))
    for m in itt.chain(load_mappings(), load_false_mappings(), load_unsure()):
        d[m.subject.prefix][m.object.prefix][m.subject.identifier] = m.object.identifier
    return {k: dict(v) for k, v in d.items()}


def prediction_tuples_from_semra(
    mappings: Iterable[semra.Mapping],
    *,
    confidence: float,
) -> list[SemanticMapping]:
    """Get prediction tuples from SeMRA mappings."""
    rows = []
    for mapping in mappings:
        try:
            row = _mapping_from_semra(mapping, confidence)
        except KeyError as e:
            tqdm.write(str(e))
            continue
        rows.append(row)
    return rows


def _mapping_from_semra(mapping: semra.Mapping, confidence: float) -> SemanticMapping:
    """Instantiate from a SeMRA mapping."""
    import pyobo
    import semra

    s_name = mapping.s.name or pyobo.get_name(mapping.s)
    if not s_name:
        raise KeyError(f"could not look up name for {mapping.s.curie}")
    o_name = mapping.o.name or pyobo.get_name(mapping.o)
    if not o_name:
        raise KeyError(f"could not look up name for {mapping.o.curie}")
    # Assume that each mapping has a single simple evidence with a mapping set annotation
    if len(mapping.evidence) != 1:
        raise ValueError
    evidence = mapping.evidence[0]
    if not isinstance(evidence, semra.SimpleEvidence):
        raise TypeError
    if evidence.mapping_set is None:
        raise ValueError
    return SemanticMapping(  # type:ignore
        subject=mapping.subject,
        predicate=mapping.predicate,
        object=mapping.object,
        mapping_justification=evidence.justification,
        confidence=confidence,
        mapping_tool=evidence.mapping_set.name,
    )



"""Web curation interface for :mod:`biomappings`."""

from __future__ import annotations

import getpass
import os
from collections import Counter, defaultdict
from collections.abc import Iterable, Iterator
from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Literal, cast, get_args

import flask
import flask_bootstrap
import pydantic
import werkzeug
from bioregistry import NormalizedNamableReference
from curies import NamableReference
from flask import current_app
from flask_wtf import FlaskForm
from pydantic import BaseModel
from typing_extensions import TypeAlias
from werkzeug.local import LocalProxy
from wtforms import StringField, SubmitField

from biomappings.resources import (
    SemanticMapping,
    append_false_mappings,
    append_true_mappings,
    append_unsure_mappings,
    load_curators,
    load_predictions,
    write_predictions,
)
from biomappings.utils import (
    BROAD_MATCH,
    EXACT_MATCH,
    MANUAL_MAPPING_CURATION,
    NARROW_MATCH,
    commit,
    get_branch,
    not_main,
    push,
)

__all__ = [
    "get_app",
]

Mark: TypeAlias = Literal["correct", "incorrect", "unsure", "broad", "narrow"]
MARKS: set[Mark] = set(get_args(Mark))


class State(BaseModel):
    """Contains the state for queries to the curation app."""

    limit: int | None = 10
    offset: int | None = 0
    query: str | None = None
    source_query: str | None = None
    source_prefix: str | None = None
    target_query: str | None = None
    target_prefix: str | None = None
    provenance: str | None = None
    prefix: str | None = None
    sort: str | None = None
    same_text: bool | None = None
    show_relations: bool = True
    show_lines: bool = False

    @classmethod
    def from_flask_globals(cls) -> State:
        """Get the state from the flask current request."""
        return State(
            limit=flask.request.args.get("limit", type=int, default=10),
            offset=flask.request.args.get("offset", type=int, default=0),
            query=flask.request.args.get("query"),
            source_query=flask.request.args.get("source_query"),
            source_prefix=flask.request.args.get("source_prefix"),
            target_query=flask.request.args.get("target_query"),
            target_prefix=flask.request.args.get("target_prefix"),
            provenance=flask.request.args.get("provenance"),
            prefix=flask.request.args.get("prefix"),
            sort=flask.request.args.get("sort"),
            same_text=_get_bool_arg("same_text"),
            show_relations=_get_bool_arg("show_relations") or current_app.config["SHOW_RELATIONS"],
            show_lines=_get_bool_arg("show_lines") or current_app.config["SHOW_LINES"],
        )


def _get_bool_arg(name: str) -> bool | None:
    value: str | None = flask.request.args.get(name, type=str)
    if value is not None:
        return value.lower() in {"true", "t"}
    return None


def url_for_state(endpoint, state: State, **kwargs: Any) -> str:
    """Get the URL for an endpoint based on the state class."""
    vv = state.model_dump(exclude_none=True, exclude_defaults=True)
    vv.update(kwargs)  # make sure stuff explicitly set overrides state
    return flask.url_for(endpoint, **vv)


def get_app(
    target_references: Iterable[NamableReference] | None = None,
    predictions_path: Path | None = None,
    positives_path: Path | None = None,
    negatives_path: Path | None = None,
    unsure_path: Path | None = None,
    controller: Controller | None = None,
) -> flask.Flask:
    """Get a curation flask app."""
    app_ = flask.Flask(__name__)
    app_.config["WTF_CSRF_ENABLED"] = False
    app_.config["SECRET_KEY"] = os.urandom(8)
    app_.config["SHOW_RELATIONS"] = True
    app_.config["SHOW_LINES"] = False
    if controller is None:
        controller = Controller(
            target_references=target_references,
            predictions_path=predictions_path,
            positives_path=positives_path,
            negatives_path=negatives_path,
            unsure_path=unsure_path,
        )
    if not controller._predictions and predictions_path is not None:
        raise RuntimeError(f"There are no predictions to curate in {predictions_path}")
    app_.config["controller"] = controller
    flask_bootstrap.Bootstrap4(app_)
    app_.register_blueprint(blueprint)
    app_.jinja_env.globals.update(
        controller=controller,
        url_for_state=url_for_state,
    )
    return app_


class Controller:
    """A module for interacting with the predictions and mappings."""

    _user: NamableReference

    def __init__(
        self,
        *,
        target_references: Iterable[NamableReference] | None = None,
        predictions_path: Path | None = None,
        positives_path: Path | None = None,
        negatives_path: Path | None = None,
        unsure_path: Path | None = None,
        user: NamableReference | None = None,
    ) -> None:
        """Instantiate the web controller.

        :param target_references:
            Pairs of prefix, local unique identifiers that are the target
            of curation. If this is given, pre-filters will be made before on predictions
            to only show ones where either the source or target appears in this set
        :param predictions_path: A custom predictions file to curate from
        :param positives_path: A custom positives file to curate to
        :param negatives_path: A custom negatives file to curate to
        :param unsure_path: A custom unsure file to curate to
        """
        self.predictions_path = predictions_path
        self._predictions = load_predictions(path=self.predictions_path)

        self.positives_path = positives_path
        self.negatives_path = negatives_path
        self.unsure_path = unsure_path

        self._marked: dict[int, Mark] = {}
        self.total_curated = 0
        self._added_mappings: list[SemanticMapping] = []
        self.target_references = set(target_references or [])

        if user is not None:
            self._current_author = user
        else:
            # FIXME this will throw an error for new user, so ask them their ORCID and name if it's missing
            self._current_author = load_curators()[getpass.getuser()]

    def _get_current_author(self) -> NamableReference:
        return self._current_author

    def predictions_from_state(self, state: State) -> Iterable[tuple[int, SemanticMapping]]:
        """Iterate over predictions from a state instance."""
        return self.predictions(
            offset=state.offset,
            limit=state.limit,
            query=state.query,
            source_query=state.source_query,
            source_prefix=state.source_prefix,
            target_query=state.target_query,
            target_prefix=state.target_prefix,
            prefix=state.prefix,
            sort=state.sort,
            same_text=state.same_text,
            provenance=state.provenance,
        )

    def predictions(
        self,
        *,
        offset: int | None = None,
        limit: int | None = None,
        query: str | None = None,
        source_query: str | None = None,
        source_prefix: str | None = None,
        target_query: str | None = None,
        target_prefix: str | None = None,
        prefix: str | None = None,
        sort: str | None = None,
        same_text: bool | None = None,
        provenance: str | None = None,
    ) -> Iterable[tuple[int, SemanticMapping]]:
        """Iterate over predictions.

        :param offset: If given, offset the iteration by this number
        :param limit: If given, only iterate this number of predictions.
        :param query: If given, show only equivalences that have it appearing as a substring in one of the source
            or target fields.
        :param source_query: If given, show only equivalences that have it appearing as a substring in one of the source
            fields.
        :param source_prefix: If given, show only mappings that have it appearing in the source prefix field
        :param target_query: If given, show only equivalences that have it appearing as a substring in one of the target
            fields.
        :param target_prefix: If given, show only mappings that have it appearing in the target prefix field
        :param prefix: If given, show only equivalences that have it appearing as a substring in one of the prefixes.
        :param same_text: If true, filter to predictions with the same label
        :param sort: If "desc", sorts in descending confidence order. If "asc", sorts in increasing confidence order.
            Otherwise, do not sort.
        :param provenance: If given, filters to provenance values matching this
        :yields: Pairs of positions and prediction dictionaries
        """
        if same_text is None:
            same_text = False
        it = self._help_it_predictions(
            query=query,
            source_query=source_query,
            source_prefix=source_prefix,
            target_query=target_query,
            target_prefix=target_prefix,
            prefix=prefix,
            sort=sort,
            same_text=same_text,
            provenance=provenance,
        )
        if offset is not None:
            try:
                for _ in range(offset):
                    next(it)
            except StopIteration:
                # if next() fails, then there are no remaining entries.
                # do not pass go, do not collect 200 euro $
                return
        if limit is None:
            yield from it
        else:
            for line_prediction, _ in zip(it, range(limit)):
                yield line_prediction

    def count_predictions_from_state(self, state: State) -> int:
        """Count the number of predictions to check for the given filters."""
        return self.count_predictions(
            query=state.query,
            source_query=state.source_query,
            source_prefix=state.source_prefix,
            target_query=state.target_query,
            target_prefix=state.target_prefix,
            prefix=state.prefix,
            same_text=state.same_text,
            provenance=state.provenance,
        )

    def count_predictions(
        self,
        query: str | None = None,
        source_query: str | None = None,
        source_prefix: str | None = None,
        target_query: str | None = None,
        target_prefix: str | None = None,
        prefix: str | None = None,
        sort: str | None = None,
        same_text: bool | None = None,
        provenance: str | None = None,
    ) -> int:
        """Count the number of predictions to check for the given filters."""
        it = self._help_it_predictions(
            query=query,
            source_query=source_query,
            source_prefix=source_prefix,
            target_query=target_query,
            target_prefix=target_prefix,
            prefix=prefix,
            sort=sort,
            same_text=same_text,
            provenance=provenance,
        )
        return sum(1 for _ in it)

    def _help_it_predictions(
        self,
        query: str | None = None,
        source_query: str | None = None,
        source_prefix: str | None = None,
        target_query: str | None = None,
        target_prefix: str | None = None,
        prefix: str | None = None,
        sort: str | None = None,
        same_text: bool | None = None,
        provenance: str | None = None,
    ) -> Iterator[tuple[int, SemanticMapping]]:
        it: Iterable[tuple[int, SemanticMapping]] = enumerate(self._predictions)
        if self.target_references:
            it = (
                (line, p)
                for (line, p) in it
                if p.subject in self.target_references or p.object in self.target_references
            )

        if query is not None:
            it = self._help_filter(
                query,
                it,
                lambda mapping: [
                    mapping.subject.curie,
                    mapping.subject.name,
                    mapping.object.curie,
                    mapping.object.name,
                    mapping.mapping_tool,
                ],
            )
        if source_prefix is not None:
            it = self._help_filter(source_prefix, it, lambda mapping: [mapping.subject.curie])
        if source_query is not None:
            it = self._help_filter(
                source_query, it, lambda mapping: [mapping.subject.curie, mapping.subject.name]
            )
        if target_query is not None:
            it = self._help_filter(
                target_query, it, lambda mapping: [mapping.object.curie, mapping.object.name]
            )
        if target_prefix is not None:
            it = self._help_filter(target_prefix, it, lambda mapping: [mapping.object.curie])
        if prefix is not None:
            it = self._help_filter(
                prefix, it, lambda mapping: [mapping.subject.curie, mapping.object.curie]
            )
        if provenance is not None:
            it = self._help_filter(provenance, it, lambda mapping: [mapping.mapping_tool])

        def _get_confidence(t: tuple[int, SemanticMapping]) -> float:
            return t[1].confidence or 0.0

        if sort is not None:
            if sort == "desc":
                it = iter(sorted(it, key=_get_confidence, reverse=True))
            elif sort == "asc":
                it = iter(sorted(it, key=_get_confidence, reverse=False))
            elif sort == "subject":
                it = iter(sorted(it, key=lambda l_p: l_p[1].subject.curie))
            elif sort == "object":
                it = iter(sorted(it, key=lambda l_p: l_p[1].object.curie))
            else:
                raise ValueError(f"unknown sort type: {sort}")

        if same_text:
            it = (
                (line, mapping)
                for line, mapping in it
                if mapping.subject.name
                and mapping.object.name
                and mapping.subject.name.casefold() == mapping.object.name.casefold()
                and mapping.predicate.curie == "skos:exactMatch"
            )

        rv = ((line, prediction) for line, prediction in it if line not in self._marked)
        return rv

    @staticmethod
    def _help_filter(
        query: str,
        it: Iterable[tuple[int, SemanticMapping]],
        func: Callable[[SemanticMapping], list[str | None]],
    ) -> Iterable[tuple[int, SemanticMapping]]:
        query = query.casefold()
        for line, mapping in it:
            if any(query in element.casefold() for element in func(mapping) if element):
                yield line, mapping

    @property
    def total_predictions(self) -> int:
        """Return the total number of yet unmarked predictions."""
        return len(self._predictions) - len(self._marked)

    def mark(self, line: int, value: Mark) -> None:
        """Mark the given equivalency as correct.

        :param line: Position of the prediction
        :param value: Value to mark the prediction with
        :raises ValueError: if an invalid value is used
        """
        if line > len(self._predictions):
            raise IndexError(
                f"given line {line} is larger than the number of predictions {len(self._predictions):,}"
            )
        if line not in self._marked:
            self.total_curated += 1
        if value not in MARKS:
            raise ValueError(f"illegal mark value given: {value}. Should be one of {MARKS}")
        self._marked[line] = value

    def add_mapping(
        self,
        subject: NormalizedNamableReference,
        obj: NormalizedNamableReference,
    ) -> None:
        """Add manually curated new mappings."""
        self._added_mappings.append(
            SemanticMapping.model_validate(
                {
                    "subject": subject,
                    "predicate": EXACT_MATCH,
                    "object": obj,
                    "author": self._get_current_author(),
                    "mapping_justification": MANUAL_MAPPING_CURATION,
                }
            )
        )
        self.total_curated += 1

    def persist(self) -> None:
        """Save the current markings to the source files."""
        if not self._marked:
            # no need to persist if there are no marks
            return None

        entries: defaultdict[Literal["correct", "incorrect", "unsure"], list[SemanticMapping]] = (
            defaultdict(list)
        )

        for line, value in sorted(self._marked.items(), reverse=True):
            try:
                mapping = self._predictions.pop(line)
            except IndexError:
                raise IndexError(
                    f"you tried popping the {line} element from the predictions list, which only has {len(self._predictions):,} elements"
                ) from None

            update: dict[str, str | NamableReference] = {
                "author": self._get_current_author(),
                "mapping_justification": MANUAL_MAPPING_CURATION,
            }

            entry_key: Literal["correct", "incorrect", "unsure"]
            # note these go backwards because of the way they are read
            if value == "broad":
                entry_key = "correct"
                update["predicate"] = NARROW_MATCH
            elif value == "narrow":
                entry_key = "correct"
                update["predicate"] = BROAD_MATCH
            elif value == "incorrect":
                entry_key = "incorrect"
                update["predicate_modifier"] = "Not"
            elif value == "correct":
                entry_key = "correct"
            elif value == "unsure":
                entry_key = "unsure"
            else:
                raise NotImplementedError

            # replace some values using model_copy since the model is frozen
            new_mapping = mapping.model_copy(update=update)

            entries[entry_key].append(new_mapping)

        # no need to standardize since we assume everything was correct on load.
        # only write files that have some valies to go in them!
        if entries["correct"]:
            append_true_mappings(
                entries["correct"], path=self.positives_path, sort=True, standardize=False
            )
        if entries["incorrect"]:
            append_false_mappings(
                entries["incorrect"], path=self.negatives_path, sort=True, standardize=False
            )
        if entries["unsure"]:
            append_unsure_mappings(
                entries["unsure"], path=self.unsure_path, sort=True, standardize=False
            )
        write_predictions(self._predictions, path=self.predictions_path)
        self._marked.clear()

        # Now add manually curated mappings, if there are any
        if self._added_mappings:
            append_true_mappings(self._added_mappings, path=self.positives_path, standardize=False)
            self._added_mappings = []


CONTROLLER: Controller = cast(Controller, LocalProxy(lambda: current_app.config["controller"]))


class MappingForm(FlaskForm):
    """Form for entering new mappings."""

    source_prefix = StringField("Source Prefix", id="source_prefix")
    source_id = StringField("Source ID", id="source_id")
    source_name = StringField("Source Name", id="source_name")
    target_prefix = StringField("Target Prefix", id="target_prefix")
    target_id = StringField("Target ID", id="target_id")
    target_name = StringField("Target Name", id="target_name")
    submit = SubmitField("Add")

    def get_subject(self) -> NormalizedNamableReference:
        """Get the subject."""
        return NormalizedNamableReference(
            prefix=self.data["source_prefix"],
            identifier=self.data["source_id"],
            name=self.data["source_name"],
        )

    def get_object(self) -> NormalizedNamableReference:
        """Get the object."""
        return NormalizedNamableReference(
            prefix=self.data["target_prefix"],
            identifier=self.data["target_id"],
            name=self.data["target_name"],
        )


blueprint = flask.Blueprint("ui", __name__)


@blueprint.route("/")
def home() -> str:
    """Serve the home page."""
    form = MappingForm()
    state = State.from_flask_globals()
    predictions = CONTROLLER.predictions_from_state(state)
    remaining_rows = CONTROLLER.count_predictions_from_state(state)
    return flask.render_template(
        "home.html",
        predictions=predictions,
        form=form,
        state=state,
        remaining_rows=remaining_rows,
    )


@blueprint.route("/summary")
def summary() -> str:
    """Serve the summary page."""
    state = State.from_flask_globals()
    state.limit = None
    predictions = CONTROLLER.predictions_from_state(state)
    counter = Counter((mapping.subject.prefix, mapping.object.prefix) for _, mapping in predictions)
    rows = []
    for (source_prefix, target_prefix), count in counter.most_common():
        row_state = deepcopy(state)
        row_state.source_prefix = source_prefix
        row_state.target_prefix = target_prefix
        rows.append((source_prefix, target_prefix, count, url_for_state(".home", row_state)))

    return flask.render_template(
        "summary.html",
        state=state,
        rows=rows,
    )


@blueprint.route("/add_mapping", methods=["POST"])
def add_mapping() -> werkzeug.Response:
    """Add a new mapping manually."""
    form = MappingForm()
    if form.is_submitted():
        try:
            subject = form.get_subject()
        except pydantic.ValidationError as e:
            flask.flash(f"Problem with source CURIE {e}", category="warning")
            return _go_home()

        try:
            obj = form.get_object()
        except pydantic.ValidationError as e:
            flask.flash(f"Problem with source CURIE {e}", category="warning")
            return _go_home()

        CONTROLLER.add_mapping(subject, obj)
        CONTROLLER.persist()
    else:
        flask.flash("missing form data", category="warning")
    return _go_home()


@blueprint.route("/commit")
def run_commit() -> werkzeug.Response:
    """Make a commit then redirect to the home page."""
    commit_info = commit(
        f"Curated {CONTROLLER.total_curated} mapping"
        f"{'s' if CONTROLLER.total_curated > 1 else ''}"
        f" ({getpass.getuser()})",
    )
    current_app.logger.warning("git commit res: %s", commit_info)
    if not_main():
        branch = get_branch()
        push_output = push(branch_name=branch)
        current_app.logger.warning("git push res: %s", push_output)
    else:
        flask.flash("did not push because on master branch")
        current_app.logger.warning("did not push because on master branch")
    CONTROLLER.total_curated = 0
    return _go_home()


CORRECT = {"yup", "true", "t", "correct", "right", "close enough", "disco"}
INCORRECT = {"no", "nope", "false", "f", "nada", "nein", "incorrect", "negative", "negatory"}
UNSURE = {"unsure", "maybe", "idk", "idgaf", "idgaff"}


def _normalize_mark(value: str) -> Mark:
    value = value.lower()
    if value in CORRECT:
        return "correct"
    elif value in INCORRECT:
        return "incorrect"
    elif value in UNSURE:
        return "unsure"
    elif value in {"broader", "broad"}:
        return "broad"
    elif value in {"narrow", "narrower"}:
        return "narrow"
    else:
        raise ValueError


@blueprint.route("/mark/<int:line>/<value>")
def mark(line: int, value: str) -> werkzeug.Response:
    """Mark the given line as correct or not."""
    CONTROLLER.mark(line, _normalize_mark(value))
    CONTROLLER.persist()
    return _go_home()


def _go_home() -> werkzeug.Response:
    state = State.from_flask_globals()
    return flask.redirect(url_for_state(".home", state))



"""Utilities for generating predictions with lexical predictions."""

from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path

import click
import pyobo
import ssslm
from bioregistry import NormalizedNamedReference, NormalizedReference
from curies import Reference
from more_click import verbose_option
from tqdm import tqdm

from biomappings import SemanticMapping
from biomappings.mapping_graph import get_mutual_mapping_filter
from biomappings.resources import append_prediction_tuples, mapping_sort_key
from biomappings.utils import EXACT_MATCH, LEXICAL_MATCHING_PROCESS, CMapping, get_script_url

__all__ = [
    "append_lexical_predictions",
    "filter_custom",
    "filter_existing_xrefs",
    "lexical_prediction_cli",
    "predict_lexical_mappings",
]

logger = logging.getLogger(__name__)


def append_lexical_predictions(
    prefix: str,
    target_prefixes: str | Iterable[str],
    provenance: str,
    *,
    relation: str | None | Reference = None,
    custom_filter: CMapping | None = None,
    identifiers_are_names: bool = False,
    path: Path | None = None,
) -> None:
    """Add lexical matching-based predictions to the Biomappings predictions.tsv file.

    :param prefix: The source prefix
    :param target_prefixes: The target prefix or prefixes
    :param provenance: The provenance text. Typically generated with ``biomappings.utils.get_script_url(__file__)``.
    :param relation: The relationship. Defaults to ``skos:exactMatch``.
    :param custom_filter: A triple nested dictionary from source prefix to target prefix to source id to target id.
        Any source prefix, target prefix, source id combinations in this dictionary will be filtered.
    :param identifiers_are_names: The source prefix's identifiers should be considered as names
    :param path: A custom path to predictions TSV file
    """
    if isinstance(target_prefixes, str):
        target_prefixes = [target_prefixes]
    # by default, PyOBO wraps a gilda grounder, but
    # can be configured to use other NER/NEN systems
    grounder = pyobo.get_grounder(target_prefixes)
    predictions = predict_lexical_mappings(
        prefix,
        predicate=relation,
        grounder=grounder,
        provenance=provenance,
        identifiers_are_names=identifiers_are_names,
    )
    if custom_filter is not None:
        predictions = filter_custom(predictions, custom_filter)
    predictions = filter_existing_xrefs(predictions, [prefix, *target_prefixes])
    predictions = sorted(predictions, key=mapping_sort_key)
    tqdm.write(f"[{prefix}] generated {len(predictions):,} predictions")
    # since the function that constructs the predictions already
    # pre-standardizes, we don't have to worry about standardizing again
    append_prediction_tuples(predictions, path=path, standardize=False)


def predict_lexical_mappings(
    prefix: str,
    provenance: str,
    *,
    predicate: str | Reference | None = None,
    grounder: ssslm.Grounder,
    identifiers_are_names: bool = False,
    strict: bool = False,
) -> Iterable[SemanticMapping]:
    """Iterate over prediction tuples for a given prefix."""
    if predicate is None:
        predicate = EXACT_MATCH
    elif isinstance(predicate, str):
        predicate = NormalizedReference.from_curie(predicate)

    id_name_mapping = pyobo.get_id_name_mapping(prefix, strict=strict)
    it = tqdm(
        id_name_mapping.items(), desc=f"[{prefix}] lexical tuples", unit_scale=True, unit="name"
    )
    name_prediction_count = 0
    for identifier, name in it:
        for scored_match in grounder.get_matches(name):
            name_prediction_count += 1
            yield SemanticMapping(
                subject=NormalizedNamedReference(prefix=prefix, identifier=identifier, name=name),
                predicate=predicate,
                object=scored_match.reference,
                mapping_justification=LEXICAL_MATCHING_PROCESS,
                confidence=round(scored_match.score, 3),
                mapping_tool=provenance,
            )

    tqdm.write(f"[{prefix}] generated {name_prediction_count:,} predictions from names")

    if identifiers_are_names:
        it = tqdm(
            pyobo.get_ids(prefix), desc=f"[{prefix}] lexical tuples", unit_scale=True, unit="id"
        )
        identifier_prediction_count = 0
        for identifier in it:
            for scored_match in grounder.get_matches(identifier):
                name_prediction_count += 1
                yield SemanticMapping(
                    subject=NormalizedNamedReference(
                        prefix=prefix, identifier=identifier, name=identifier
                    ),
                    predicate=predicate,
                    object=scored_match.reference,
                    mapping_justification=LEXICAL_MATCHING_PROCESS,
                    confidence=round(scored_match.score, 3),
                    mapping_tool=provenance,
                )
        tqdm.write(
            f"[{prefix}] generated {identifier_prediction_count:,} predictions from identifiers"
        )


def filter_custom(
    mappings: Iterable[SemanticMapping],
    custom_filter: CMapping,
) -> Iterable[SemanticMapping]:
    """Filter out custom mappings."""
    counter = 0
    for mapping in mappings:
        if (
            custom_filter.get(mapping.subject.prefix, {})
            .get(mapping.object.prefix, {})
            .get(mapping.subject.identifier)
        ):
            counter += 1
            continue
        yield mapping
    logger.info("filtered out %d custom mapped matches", counter)


def filter_existing_xrefs(
    mappings: Iterable[SemanticMapping], prefixes: Iterable[str]
) -> Iterable[SemanticMapping]:
    """Filter predictions that match xrefs already loaded through PyOBO."""
    prefixes = set(prefixes)

    entity_to_mapped_prefixes: defaultdict[Reference, set[str]] = defaultdict(set)
    for subject_prefix in prefixes:
        for subject_id, target_prefix, object_id in pyobo.get_xrefs_df(subject_prefix).values:
            entity_to_mapped_prefixes[
                NormalizedReference(prefix=subject_prefix, identifier=subject_id)
            ].add(target_prefix)
            entity_to_mapped_prefixes[
                NormalizedReference(prefix=target_prefix, identifier=object_id)
            ].add(subject_prefix)

    n_predictions = 0
    for mapping in tqdm(mappings, desc="filtering predictions"):
        if (
            mapping.object.prefix in entity_to_mapped_prefixes[mapping.subject]
            or mapping.subject.prefix in entity_to_mapped_prefixes[mapping.object]
        ):
            n_predictions += 1
            continue
        yield mapping

    tqdm.write(
        f"filtered out {n_predictions:,} pre-mapped matches",
    )


def lexical_prediction_cli(
    script: str,
    prefix: str,
    target: str | list[str],
    *,
    filter_mutual_mappings: bool = False,
    identifiers_are_names: bool = False,
) -> None:
    """Construct a CLI and run it."""
    tt = target if isinstance(target, str) else ", ".join(target)

    @click.command(help=f"Generate mappings from {prefix} to {tt}")
    @verbose_option
    def main() -> None:
        """Generate mappings."""
        if filter_mutual_mappings:
            mutual_mapping_filter = get_mutual_mapping_filter(prefix, target)
        else:
            mutual_mapping_filter = None

        append_lexical_predictions(
            prefix,
            target,
            custom_filter=mutual_mapping_filter,
            provenance=get_script_url(script),
            identifiers_are_names=identifiers_are_names,
        )

    main()
